/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 2014-2016 Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

load "s7-1200-basic".
load "lib-conv".

       
/* the following ladder logic 
        PL3     PL0
    |---| |-----( )---|
    |                 |
    |   PL2     PL1   |
    |---| |-----( )---|
    |                 |
    |   PB      PL2   |
    |---| |-----( )---|
    |                 |
    |   PL1     PL3   |
    |---| |-----( )---|
    |                 |
    is translated into Tempura using the constructs defined in s7-1200-basic.t
    
 */
define plc(PB,PL) = {
  /* let's output the value of PB and PL in each state */
  always ( format("PB=%d, PL=%t\n",PB,PL) )  and
  { /* the first rung */
   { COIL(NO(PL[3]),PL[0]) and skip and no_ch(PB,PL[1],PL[2],PL[3]) };
    /* the second rung */
   { COIL(NO(PL[2]),PL[1]) and skip and no_ch(PB,PL[0],PL[2],PL[3]) };
    /* the third rung */
   { COIL(NO(PB),PL[2])    and skip and no_ch(PB,PL[0],PL[1],PL[3]) };
    /* the fourth rung */
   { COIL(NO(PL[1]),PL[3]) and skip and no_ch(PB,PL[0],PL[1],PL[2]) }
  }
}.

/* X : used to store date read from external file
   I0: input  0 to the plc is 8 bits
   Q0: output 0 from the plc is 8 bits
   PL: expected output is 4 bits
*/   
define test_plc(F) = { 
   exists X,I0,Q0,PL : {
    /* F is the file we process data from */
    set infile=F and 
    /* read the first data element from the file */	   
    input X  and
    /* PL is 4 bits, I0 is 8 bits, Q0 is 8 bits */
    list(PL,4) and stable(struct(PL)) and
    list(I0,8) and stable(struct(I0)) and
    list(Q0,8) and stable(struct(Q0)) and
    /* I0, Q0 and PL are initially all zeros */
    I0=[0,0,0,0,0,0,0,0] and Q0=[0,0,0,0,0,0,0,0] and
    PL=[0,0,0,0] and
    /* as long the input is not zero we continue otherwise we stop */
    while (X ~= 0) do {
       /* we discard the first input as it is MK: data */
       {skip and stable(I0) and stable(Q0) and stable(PL)};
       /* the second data element is input to the plc and stored in I0 */
       { input X and output(X) and skip and
         if X[0..3]="DI:" then {
          I0:=inttobits(hexbinbyte(X[3],X[4]))
         } and
         stable(Q0) and stable(PL)
       };
       /* the third data element is output from the plc and stored in Q0 */
       { input X and output(X) and skip and
         if X[0..3]="DO:" then {
          Q0:=inttobits(hexbinbyte(X[3],X[4]))
         } and
         stable(I0) and stable(PL)
       }; /* the fourth data element is CT data and discarded */
       { input X and skip and
         stable(I0) and stable(Q0) and stable(PL)
       }; /* the fifth data element is TI data and discarded */
       { input X  and skip and
         stable(I0) and stable(Q0) and stable(PL)
       }; /* we feed the input I0 to our formal specified plc and store
             the output in PL */
       { plc(I0[0],PL) and stable(I0) and stable(Q0) 
       }; /* we start all over again but also check whether the
             received output Q0 matches the output PL from the
	     formal specification */
       { input X  and empty and
         forall i<4 : {
          if PL[i] ~= Q0[i] then {
            format("Output Q0[%d]=%t does not match expected output PL[%d]=%t\n",i,Q0[i],i,PL[i])
	  }
	 }
       }	  
    }
  }
}.

/* run */ define test_correct() = { test_plc("S71200-correct.TXT") }.

/* run */ define test_faulty() = { test_plc("S71200-faulty.TXT") }.
