/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2016 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: gc.c 273 2016-03-18 10:59:02Z cau $
 */


#include "system.h"

/**********************\
*                      *
*  Garbage collector   *
*                      *
\**********************/
void collect_garbage()
{
    if (store_verbosity) {
        fputs ("***Tempura: garbage collection: ",stdout);
    }
    mark_phase();
    reclaim_phase();
    if (store_verbosity) {
                printf ("done (%d%% in use).\n",
		HeapSize == 0 ? 100 : ((HeapSize-FreeNodes)*100)/HeapSize);
                fflush(stdout);
    }
}

void mark_phase()
{
    REDUCESTACKPTR rs;
    NODEPTR np;
    if (store_verbosity)
        fputs ("mark, ", stdout);
    for (rs = ReduceStack; rs <= ReduceStackTop; rs += 1) {
        mark (rs -> n);
    }
    for (np = TempNodeList; np != NULL; np = node_next(np)) {
        mark (np);
    }
    mark (node_env_idlist(EnvStack[EnvBase].env)); 
    mark_mem (GlobalMemory, GlobalMemUsed);

        mark (Parseroot);
    
    mark (LastCommand);
    mark (SysWaitList);
    mark (iRememberList);
    /* Following two added by BCM.  14Dec96 */
    mark(Path);
    mark(HelpDir);
    
    mark(OldInFile);
    mark(InFile);
    mark(OutFile);
    
}

void mark_mem(loc,n)
NODEPTR loc;
unsigned n;
{
    unsigned int i;
    
    while (n-- > 0) {
        loc = curr_loc (loc);
        node_mark(loc) = T;
        if (is_indirect_mem (loc))
            mark_mem (node_next(loc), 1);
        else {
            if (mem_defined1 (loc))
                mark (node_next(loc));
            loc = next_loc (loc);
            node_mark(loc) = T;
            if (mem_defined1 (loc))
                mark (node_next(loc));
            for (i=0; i<=1+node_mem_history(loc); i = i + 1) {
                loc = prev_loc1 (loc,i);
                node_mark(loc) = T;
                if (mem_defined1 (loc))
                    mark(node_next(loc));
            }
                /*loc = prev_loc (loc);
            if (mem_defined (loc))
            mark (node_next(loc));*/
	}
        loc += offset2(1,node_mem_history(loc));
    }
}

void mark(e)
NODEPTR e;
{
    if (e == NULL)
        return;
    if (node_mark(e))
        return;
    node_mark(e) = T;
    switch (node_type(e)) {
        case STATE:
        case STATIC:
            mark (node_next(e));
            break;
        case ENV:
            mark (node_env_idlist(e)); 
            break;
        case BND:
	    mark (node_bnd_env(e)); 
	    mark (node_bnd_blk(e)); 
	    break;
        case NAME:
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
            break;
        case LOC:
            mark_mem (node_loc_ptr(e), 1);
            break;
        case LIST:
            mark (node_lst_blk(e)); 
            break;
        case MEMBLOCK:
            mark_mem (node_blk_base(e), node_blk_size(e));
            break;
        case EXISTS:
            mark (node_lnk_4(e));
            mark (node_lnk_5(e));
        case FOR:
        case FORALL:
        case IF:
        case SUBLIST:
        case FORIN:
        case iAND:
        case iCALL:
        case iPROCESS:
        case iPREFIX:
        case iPROJ:
        case iSEMI:
        case iFOR:
        case iFORIN:
        case iIF:    
            mark (node_lnk_3(e));
        case iOR:
        case THE:
        case FORTIMES:
        case iEXISTS:
        case PROJ:
        case COMMA:
        case DEFINE:
        case LAMBDA:
        case WHILE:
        case REPEAT:
        case LT:
        case LE:
        case EQ:
        case GE:
        case GT:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case SEMI:
        case WHERE:
        case AND:
        case OR:
        case IMPLIES:
        case GETS:
        case IS:
        case MOD:
        case ATANTWO:
        case ASSN:
        case UASSN:
        case SKIPUASSN:   
        case CALL:
        case SUB:
        case FIXLIST:
        case VARLIST:
        case NPREV:
        case RUN:
        case SET:
        case SHOWMEM:
        case DEFAULTS:
        case iSYSTEM:
            mark (node_lnk_2(e));
        case NOT:
        case UPLUS:
        case UMINUS:
        case CEIL:
        case FLOOR:
        case SQRT:
        case ITOF:
        case EXPO:
        case LOG:
        case LOGTEN:
        case SIN:
        case COS:
        case TAN:
        case ASIN:
        case ACOS:
        case ATAN:
        case SINH:
        case COSH:
        case TANH:
        case FABS:
        case PROCESS:
        case PREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case NEXT:
        case WNEXT:
        case PREV:
        case LEN:
        case STB:
        case MEM:
        case STRUCT:
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case PEEK:
        case HELP:
        case TYPE:
        case iUNDEFINED:
        case iREMEMBER:
        case ASCII:
        case FORMAT:
        case SIZE:
        case MAKELIST:
        case REFER:
        case DEREF:
        case LOAD:
        case SYSTEM:
        case iLEN:    
        case iFORTIMES:
            mark (node_lnk_1(e));
            break;
        case END:
        case SKIP:
        case EMPTY:
        case MORE:
        case RANLEN:
        case INFINITE:    
        case EXIT:
        case SHOWTRACE:
        case ABORT:
        case BREAK:
        case RETURN:
        case CONTINUE:
        case LEAP:
        case STEP:
        case SHOWSTATE:
	case PASSLIMIT:
	case VERBOSITY:
	case PRINTLEVEL:
	case DEBUGLEVEL:
	case STATEPRINT:
	case REDUNDANCY:
	case CALLBYNAME:
	case ASSIGNAHEAD:
        case BREAKISABORT:
        case HOPCHOP:
	case PATH:
	case HELPDIR:
        case INFILE:
        case OUTFILE:
	case TIMESTEP:
        case STATENUMBER:
        case RANDOM:    
        case RANDSEED:
	case RRANDOM:    
                /*case RRANDSEED:*/
        case SRANDOM:    
                /*case SRANDSEED:*/
	case SRRANDOM:    
                /*case SRRANDSEED:*/    
        case MAXRANDLEN:
        case PRECISION:
            break;
        default:
            error_break1d ("mark: unknown node: %d", node_type(e));
            break;
    }
}


void reclaim_phase()
{
    HEAP h;
    NODEPTR np,base,top;
    if (store_verbosity)
        fputs("reclaim, ",stdout);
    FreeNodeList = NULL;
    FreeNodes = 0;
    for (h = Heap; h != NULL; h = h -> next_block) {
        base = h -> base;
	top = h -> top;
	for (np = base; np < top; np++) {                 
            if ((node_type(np) == MEMBLOCK) && is_refd_block (np))
                node_mark(np) = F;
            else if (node_mark(np))
                node_mark(np) = F;
            else 
                reclaim_node (np);
	}
    }
}


int is_refd_block(bp)
NODEPTR bp;
{
    int r = F;
    register NODEPTR loc = node_blk_base(bp);
    register int n = node_blk_size(bp);
    register unsigned int i;
    
    while (n-- > 0) {
        loc = curr_loc (loc);
        if (node_mark(loc)) {
            r = T;
            node_mark(loc) = F;
        }
        loc = next_loc (loc);
        if (node_mark(loc)) {
            r = T;
            node_mark(loc) = F;
        }
        for (i=0; i<=1+node_mem_history(loc); i = i + 1) {
            loc = prev_loc1 (loc,i);
            if (node_mark(loc)) {
                r = T;
                node_mark(loc) = F;
            }
        }
               
        loc += offset2(1,node_mem_history(loc));
    }
    return (r);
}

void reclaim_node(np)
NODEPTR np;
{
    switch (node_type(np)) {
        case STRING:
                free_mem (node_strval(np));
            break;
        case MEMBLOCK:
	    LocalMemUsed -= node_blk_size(np);
            free_mem (node_blk_base(np));
            break;
        default:
            break;
    }
    node_type(np) = 0;
    node_next(np) = FreeNodeList;
    FreeNodeList = np;
        /*if (FreeNodes < HeapSize)*/
    FreeNodes += 1;
        
}



